@@ -39,13 +39,15 @@ ClusterShell library which implements some features of the NodeSet
 and RangeSet classes.
 """
 
+import math
 import sys
 
 from ClusterShell.CLI.Error import GENERIC_ERRORS, handle_generic_error
 from ClusterShell.CLI.OptionParser import OptionParser
 from ClusterShell.CLI.Utils import NodeSet  # safe import
 
-from ClusterShell.NodeSet import RangeSet, grouplist, RESOLVER_STD_GROUP
+from ClusterShell.NodeSet import RangeSet, grouplist, std_group_resolver
+from ClusterShell.NodeUtils import GroupSourceNoUpcall
 
 
 def process_stdin(xsetop, xsetcls, autostep):
@@ -96,40 +98,63 @@ def compute_nodeset(xset, args, autostep):
 
     return xset
 
-def command_list(options, xset):
-    """List (-l/-ll/-lll) command handler."""
-    list_level = options.list
+def print_source_groups(source, level, xset, opts):
+    """
+    Print groups from a source, a level of verbosity and an optional
+    nodeset acting as a filter.
+    """
     # list groups of some specified nodes?
-    if options.all or xset or \
-        options.and_nodes or options.sub_nodes or options.xor_nodes:
+    if opts.all or xset or opts.and_nodes or opts.sub_nodes or opts.xor_nodes:
         # When some node sets are provided as argument, the list command
         # retrieves node groups these nodes belong to, thanks to the
-        # groups() method (new in 1.6). Note: stdin support is enabled
-        # when the '-' special character is encountered.
-        groups = xset.groups(options.groupsource, options.groupbase)
+        # groups() method.
+        # Note: stdin support is enabled when '-' is found.
+        groups = xset.groups(source, opts.groupbase)
         for group, (gnodes, inodes) in groups.iteritems():
-            if list_level == 1:         # -l
+            if level == 1:
                 print group
-            elif list_level == 2:       # -ll
+            elif level == 2:
                 print "%s %s" % (group, inodes)
-            else:                       # -lll
-                print "%s %s %d/%d" % (group, inodes, len(inodes), \
+            else:
+                print "%s %s %d/%d" % (group, inodes, len(inodes),
                                        len(gnodes))
-        return
-    # "raw" group list when no argument at all
-    for group in grouplist(options.groupsource):
-        if options.groupsource and not options.groupbase:
-            nsgroup = "@%s:%s" % (options.groupsource, group)
-        else:
-            nsgroup = "@%s" % group
-        if list_level == 1:         # -l
-            print nsgroup
-        else:
-            nodes = NodeSet(nsgroup)
-            if list_level == 2:     # -ll
-                print "%s %s" % (nsgroup, nodes)
-            else:                   # -lll
-                print "%s %s %d" % (nsgroup, nodes, len(nodes))
+    else:
+        # "raw" group list when no argument at all
+        for group in grouplist(source):
+            if source and not opts.groupbase:
+                nsgroup = "@%s:%s" % (source, group)
+            else:
+                nsgroup = "@%s" % group
+            if level == 1:
+                print nsgroup
+            else:
+                nodes = NodeSet(nsgroup)
+                if level == 2:
+                    print "%s %s" % (nsgroup, nodes)
+                else:
+                    print "%s %s %d" % (nsgroup, nodes, len(nodes))
+
+def command_list(options, xset, group_resolver):
+    """List command handler (-l/-ll/-lll/-L/-LL/-LLL)."""
+    list_level = options.list + options.listall
+    if options.listall:
+        # useful: sources[0] is always the default or selected source
+        sources = group_resolver.sources()
+        # do not print name of default group source unless -s specified
+        if not options.groupsource:
+            sources[0] = None
+    else:
+        sources = [options.groupsource]
+
+    for source in sources:
+        try:
+            print_source_groups(source, list_level, xset, options)
+        except GroupSourceNoUpcall, exc:
+            if not options.listall:
+                raise
+            # missing list upcall is not fatal with -L
+            msgfmt = "Warning: No %s upcall defined for group source %s"
+            print >>sys.stderr, msgfmt % (exc, source)
 
 def nodeset():
     """script subroutine"""
@@ -142,13 +167,16 @@ def nodeset():
     parser.install_nodeset_options()
     (options, args) = parser.parse_args()
 
+    group_resolver = std_group_resolver()
+
     if options.debug:
-        RESOLVER_STD_GROUP.set_verbosity(1)
+        group_resolver.set_verbosity(1)
 
     # Check for command presence
     cmdcount = int(options.count) + int(options.expand) + \
                int(options.fold) + int(bool(options.list)) + \
-               int(options.regroup) + int(options.groupsources)
+               int(bool(options.listall)) + int(options.regroup) + \
+               int(options.groupsources)
     if not cmdcount:
         parser.error("No command specified.")
     elif cmdcount > 1:
@@ -167,66 +195,72 @@ def nodeset():
     if options.maxsplit is None:
         options.maxsplit = 1
 
-    if options.groupsource and not options.quiet and \
-       (class_set == RangeSet or options.groupsources):
+    if options.axis and (not options.fold or options.rangeset):
+        parser.error("--axis option is only supported when folding nodeset")
+
+    if options.groupsource and not options.quiet and class_set == RangeSet:
         print >> sys.stderr, "WARNING: option group source \"%s\" ignored" \
                                 % options.groupsource
 
+    # We want -s <groupsource> to act as a substition of default groupsource
+    # (ie. it's not necessary to prefix group names by this group source).
+    if options.groupsource:
+        group_resolver.default_source_name = options.groupsource
+
     # The groupsources command simply lists group sources.
     if options.groupsources:
         if options.quiet:
             dispdefault = ""    # don't show (default) if quiet is set
         else:
             dispdefault = " (default)"
-        for src in RESOLVER_STD_GROUP.sources():
+        for src in group_resolver.sources():
             print "%s%s" % (src, dispdefault)
             dispdefault = ""
         return
 
-    # We want -s <groupsource> to act as a substition of default groupsource
-    # (ie. it's not necessary to prefix group names by this group source).
-    if options.groupsource:
-        RESOLVER_STD_GROUP.default_sourcename = options.groupsource
+    autostep = options.autostep
+
+    # Do not use autostep for computation when a percentage or the special
+    # value 'auto' is specified. Real autostep value is set post-process.
+    if type(autostep) is float or autostep == 'auto':
+        autostep = None
 
     # Instantiate RangeSet or NodeSet object
-    xset = class_set(autostep=options.autostep)
+    xset = class_set(autostep=autostep)
 
     if options.all:
         # Include all nodes from external node groups support.
-        xset.update(NodeSet.fromall()) # uses default_sourcename
+        xset.update(NodeSet.fromall()) # uses default_source when set
 
-    if not args and not options.all and not options.list:
+    if not args and not options.all and not (options.list or options.listall):
         # No need to specify '-' to read stdin in these cases
-        process_stdin(xset.update, xset.__class__, options.autostep)
+        process_stdin(xset.update, xset.__class__, autostep)
 
     # Apply first operations (before first non-option)
     for nodes in options.and_nodes:
         if nodes == '-':
-            process_stdin(xset.intersection_update, xset.__class__,
-                          options.autostep)
+            process_stdin(xset.intersection_update, xset.__class__, autostep)
         else:
-            xset.intersection_update(class_set(nodes,
-                                               autostep=options.autostep))
+            xset.intersection_update(class_set(nodes, autostep=autostep))
     for nodes in options.sub_nodes:
         if nodes == '-':
-            process_stdin(xset.difference_update, xset.__class__,
-                          options.autostep)
+            process_stdin(xset.difference_update, xset.__class__, autostep)
         else:
-            xset.difference_update(class_set(nodes, autostep=options.autostep))
+            xset.difference_update(class_set(nodes, autostep=autostep))
     for nodes in options.xor_nodes:
         if nodes == '-':
             process_stdin(xset.symmetric_difference_update, xset.__class__,
-                          options.autostep)
+                          autostep)
         else:
-            xset.symmetric_difference_update(class_set(nodes, \
-                                             autostep=options.autostep))
+            xset.symmetric_difference_update(class_set(nodes,
+                                                       autostep=autostep))
 
     # Finish xset computing from args
-    compute_nodeset(xset, args, options.autostep)
+    compute_nodeset(xset, args, autostep)
 
     # The list command has a special handling
-    if options.list > 0:
-        return command_list(options, xset)
+    if options.list > 0 or options.listall > 0:
+        return command_list(options, xset, group_resolver)
 
     # Interprete special characters (may raise SyntaxError)
     separator = eval('\'%s\'' % options.separator, {"__builtins__":None}, {})
@@ -237,16 +271,36 @@ def nodeset():
             _xset.update(xset[sli])
         xset = _xset
 
+    if options.autostep == 'auto':
+        # Simple implementation of --autostep=auto
+        # if we have at least 3 nodes, all index should be foldable as a-b/n
+        xset.autostep = max(3, len(xset))
+    elif type(options.autostep) is float:
+        # at least % of nodes should be foldable as a-b/n
+        autofactor = float(options.autostep)
+        xset.autostep = int(math.ceil(float(len(xset)) * autofactor))
+
+    # user-specified nD-nodeset fold axis
+    if options.axis:
+        if not options.axis.startswith('-'):
+            # axis are 1-indexed in nodeset CLI (0 ignored)
+            xset.fold_axis = tuple(x - 1 for x in RangeSet(options.axis) if x > 0)
+        else:
+            # negative axis index (only single number supported)
+            xset.fold_axis = [int(options.axis)]
+
+    fmt = options.output_format # default to '%s'
+
     # Display result according to command choice
     if options.expand:
-        xsubres = lambda x: separator.join(x.striter())
+        xsubres = lambda x: separator.join((fmt % s for s in x.striter()))
     elif options.fold:
-        xsubres = lambda x: x
+        xsubres = lambda x: fmt % x
     elif options.regroup:
-        xsubres = lambda x: x.regroup(options.groupsource, \
-                                      noprefix=options.groupbase)
+        xsubres = lambda x: fmt % x.regroup(options.groupsource,
+                                            noprefix=options.groupbase)
     else:
-        xsubres = len
+        xsubres = lambda x: fmt % len(x)
 
     if not xset or options.maxsplit <= 1 and not options.contiguous:
         print xsubres(xset)
@@ -262,12 +316,9 @@ def main():
     """main script function"""
     try:
         nodeset()
-    except AssertionError, ex:
+    except (AssertionError, IndexError, ValueError), ex:
         print >> sys.stderr, "ERROR:", ex
         sys.exit(1)
-    except IndexError:
-        print >> sys.stderr, "ERROR: syntax error"
-        sys.exit(1)
     except SyntaxError:
         print >> sys.stderr, "ERROR: invalid separator"
         sys.exit(1)
