@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 #
-# Copyright CEA/DAM/DIF (2010, 2011, 2012)
-#  Contributor: Stephane THIELL <stephane.thiell@cea.fr>
+# Copyright CEA/DAM/DIF (2010-2015)
+#  Contributor: Stephane THIELL <sthiell@stanford.edu>
 #
 # This file is part of the ClusterShell library.
 #
@@ -32,7 +32,7 @@
 # knowledge of the CeCILL-C license and that you accept its terms.
 
 """
-common ClusterShell CLI OptionParser
+Common ClusterShell CLI OptionParser
 
 With few exceptions, ClusterShell command-lines share most option
 arguments. This module provides a common OptionParser class.
@@ -45,6 +45,20 @@ from ClusterShell import __version__
 from ClusterShell.Engine.Factory import PreferredEngine
 from ClusterShell.CLI.Display import THREE_CHOICES
 
+
+def check_autostep(option, opt, value):
+    """type-checker function for autostep"""
+    try:
+        if '%' in value:
+            return float(value[:-1]) / 100.0
+        return int(value)
+    except ValueError:
+        if value == 'auto':
+            return value
+        error_fmt = "option %s: invalid value: %r, should be node count, " \
+                    "node percentage or 'auto'"
+        raise optparse.OptionValueError(error_fmt % (opt, value))
+
 def check_safestring(option, opt, value):
     """type-checker function for safestring"""
     try:
@@ -60,13 +74,14 @@ def check_safestring(option, opt, value):
 
 class Option(optparse.Option):
     """This Option subclass adds a new safestring type."""
-    TYPES = optparse.Option.TYPES + ("safestring",)
+    TYPES = optparse.Option.TYPES + ("autostep", "safestring",)
     TYPE_CHECKER = copy(optparse.Option.TYPE_CHECKER)
+    TYPE_CHECKER["autostep"] = check_autostep
     TYPE_CHECKER["safestring"] = check_safestring
 
 class OptionParser(optparse.OptionParser):
     """Derived OptionParser for all CLIs"""
-    
+
     def __init__(self, usage, **kwargs):
         """Initialize ClusterShell CLI OptionParser"""
         optparse.OptionParser.__init__(self, usage,
@@ -82,6 +97,12 @@ class OptionParser(optparse.OptionParser):
                         type="safestring", dest="groupsource",
                         help="optional groups.conf(5) group source to use")
 
+    def install_config_options(self, filename=''):
+        """Install config options override"""
+        self.add_option("-O", "--option", action="append", metavar="KEY=VALUE",
+                        dest="option", default=[],
+                        help="override any key=value %s options" % filename)
+
     def install_nodes_options(self):
         """Install nodes selection options"""
         optgrp = optparse.OptionGroup(self, "Selecting target nodes")
@@ -100,8 +121,13 @@ class OptionParser(optparse.OptionParser):
         optgrp.add_option("-E", "--engine", action="store", dest="engine",
                           choices=["auto"] + PreferredEngine.engines.keys(),
                           default="auto", help=optparse.SUPPRESS_HELP)
-        optgrp.add_option("-T", "--topology", action="store", dest="topofile",
-                          default=None, help=optparse.SUPPRESS_HELP)
+        optgrp.add_option("--hostfile", "--machinefile", action="append",
+                          dest="hostfile", default=[], metavar='FILE',
+                          help="path to file containing a list of target hosts")
+        optgrp.add_option("--topology", action="store", dest="topofile",
+                          default=None, metavar='FILE',
+                          help="topology configuration file to use for tree "
+                               "mode")
         self.add_option_group(optgrp)
 
     def install_display_options(self,
@@ -132,6 +158,8 @@ class OptionParser(optparse.OptionParser):
             optgrp.add_option("-b", "-c", "--dshbak", action="store_true",
                 dest="gather", help="gather nodes with same output")
         else:
+            optgrp.add_option("-P", "--progress", action="store_true",
+                dest="progress", help="show progress during command execution")
             optgrp.add_option("-b", "--dshbak", action="store_true",
                 dest="gather", help="gather nodes with same output")
         optgrp.add_option("-B", action="store_true", dest="gatherall",
@@ -143,8 +171,8 @@ class OptionParser(optparse.OptionParser):
         if separator_option:
             optgrp.add_option("-S", "--separator", action="store",
                               dest="separator", default=':',
-                              help="node / line content separator string " \
-                              "(default: ':')")
+                              help="node / line content separator string "
+                                   "(default: ':')")
         else:
             optgrp.add_option("-S", action="store_true", dest="maxrc",
                               help="return the largest of command return codes")
@@ -159,11 +187,11 @@ class OptionParser(optparse.OptionParser):
                               help="message tree trace mode")
             optgrp.add_option("--interpret-keys", action="store",
                               dest="interpret_keys", choices=THREE_CHOICES,
-                              default=THREE_CHOICES[-1], help="whether to " \
+                              default=THREE_CHOICES[-1], help="whether to "
                               "interpret keys (never, always or auto)")
 
         optgrp.add_option("--color", action="store", dest="whencolor",
-                          choices=THREE_CHOICES, help="whether to use ANSI " \
+                          choices=THREE_CHOICES, help="whether to use ANSI "
                           "colors (never, always or auto)")
         optgrp.add_option("--diff", action="store_true", dest="diff",
                           help="show diff between gathered outputs")
@@ -191,48 +219,65 @@ class OptionParser(optparse.OptionParser):
                           help="preserve modification times and modes")
         self.add_option_group(optgrp)
 
-    
-    def install_ssh_options(self):
-        """Install engine/connector (ssh) options"""
-        optgrp = optparse.OptionGroup(self, "Ssh/Tree options")
-        optgrp.add_option("-f", "--fanout", action="store", dest="fanout", 
+
+    def install_connector_options(self):
+        """Install engine/connector (ssh, ...) options"""
+        optgrp = optparse.OptionGroup(self, "Connection options")
+        optgrp.add_option("-f", "--fanout", action="store", dest="fanout",
                           help="use a specified fanout", type="int")
         #help="queueing delay for traffic grooming"
-        optgrp.add_option("-Q", action="store", dest="grooming_delay", 
+        optgrp.add_option("--grooming", action="store", dest="grooming_delay",
                           help=optparse.SUPPRESS_HELP, type="float")
         optgrp.add_option("-l", "--user", action="store", type="safestring",
                           dest="user", help="execute remote command as user")
         optgrp.add_option("-o", "--options", action="store", dest="options",
                           help="can be used to give ssh options")
         optgrp.add_option("-t", "--connect_timeout", action="store",
-                          dest="connect_timeout", help="limit time to " \
-                          "connect to a node" ,type="float")
+                          dest="connect_timeout",
+                          help="limit time to connect to a node", type="float")
         optgrp.add_option("-u", "--command_timeout", action="store",
-                          dest="command_timeout", help="limit time for " \
-                          "command to run on the node", type="float")
+                          dest="command_timeout",
+                          help="limit time for command to run on the node",
+                          type="float")
+        optgrp.add_option("-R", "--worker", action="store", dest="worker",
+                          help="worker name to use for command execution "
+                               "('exec', 'rsh', 'ssh', etc. default is 'ssh')")
+        optgrp.add_option("--remote", action="store", dest="remote",
+                          choices=('yes', 'no'),
+                          help="whether to enable remote execution: in tree "
+                               "mode, 'yes' forces connections to the leaf "
+                               "nodes for execution, 'no' establishes "
+                               "connections up to the leaf parent nodes for "
+                               "execution (default is 'yes')")
         self.add_option_group(optgrp)
 
     def install_nodeset_commands(self):
         """Install nodeset commands"""
         optgrp = optparse.OptionGroup(self, "Commands")
-        optgrp.add_option("-c", "--count", action="store_true", dest="count", 
-                          default=False, help="show number of nodes in " \
-                          "nodeset(s)")
-        optgrp.add_option("-e", "--expand", action="store_true",
-                          dest="expand", default=False, help="expand " \
-                          "nodeset(s) to separate nodes")
-        optgrp.add_option("-f", "--fold", action="store_true", dest="fold", 
-                          default=False, help="fold nodeset(s) (or " \
-                          "separate nodes) into one nodeset")
-        optgrp.add_option("-l", "--list", action="count", dest="list", 
-                          default=False, help="list node groups (see -s " \
-                          "GROUPSOURCE)")
+        optgrp.add_option("-c", "--count", action="store_true", dest="count",
+                          default=False,
+                          help="show number of nodes in nodeset(s)")
+        optgrp.add_option("-e", "--expand", action="store_true", dest="expand",
+                          default=False,
+                          help="expand nodeset(s) to separate nodes")
+        optgrp.add_option("-f", "--fold", action="store_true", dest="fold",
+                          default=False, help="fold nodeset(s) (or separate "
+                                              "nodes) into one nodeset")
+        optgrp.add_option("-l", "--list", action="count", dest="list",
+                          default=False, help="list node groups from one "
+                                              "source (see -s GROUPSOURCE)")
+        optgrp.add_option("-L", "--list-all", action="count", dest="listall",
+                          default=False,
+                          help="list node groups from all group sources")
         optgrp.add_option("-r", "--regroup", action="store_true",
-                          dest="regroup", default=False, help="fold nodes " \
-                          "using node groups (see -s GROUPSOURCE)")
-        optgrp.add_option("--groupsources", action="store_true",
-                          dest="groupsources", default=False, help="list " \
-                          "all configured group sources (see groups.conf(5))")
+                          dest="regroup", default=False,
+                          help="fold nodes using node groups (see -s "
+                               "GROUPSOURCE)")
+        optgrp.add_option("--list-sources", "--groupsources",
+                          action="store_true", dest="groupsources",
+                          default=False,
+                          help="list all active group sources (see "
+                               "groups.conf(5))")
         self.add_option_group(optgrp)
 
     def install_nodeset_operations(self):
@@ -245,42 +290,49 @@ class OptionParser(optparse.OptionParser):
                           dest="and_nodes", default=[], type="string",
                           help="calculate nodesets intersection")
         optgrp.add_option("-X", "--xor", action="append", dest="xor_nodes",
-                          default=[], type="string", help="calculate " \
-                          "symmetric difference between nodesets")
+                          default=[], type="string",
+                          help="calculate symmetric difference between "
+                               "nodesets")
         self.add_option_group(optgrp)
 
     def install_nodeset_options(self):
         """Install nodeset options"""
         optgrp = optparse.OptionGroup(self, "Options")
-        optgrp.add_option("-a", "--all", action="store_true", dest="all", 
-                          help="call external node groups support to " \
+        optgrp.add_option("-a", "--all", action="store_true", dest="all",
+                          help="call external node groups support to "
                                "display all nodes")
-        optgrp.add_option("--autostep", action="store", dest="autostep", 
-                          help="auto step threshold number when folding " \
-                               "nodesets", type="int")
+        optgrp.add_option("--autostep", action="store", dest="autostep",
+                          help="enable a-b/step style syntax when folding, "
+                               "value is min node count threshold (eg. '4', "
+                               "'50%' or 'auto')",
+                          type="autostep")
         optgrp.add_option("-d", "--debug", action="store_true", dest="debug",
                           help="output more messages for debugging purpose")
         optgrp.add_option("-q", "--quiet", action="store_true", dest="quiet",
                           help="be quiet, print essential output only")
         optgrp.add_option("-R", "--rangeset", action="store_true",
-                          dest="rangeset", help="switch to RangeSet instead " \
-                          "of NodeSet. Useful when working on numerical " \
+                          dest="rangeset", help="switch to RangeSet instead "
+                          "of NodeSet. Useful when working on numerical "
                           "cluster ranges, eg. 1,5,18-31")
         optgrp.add_option("-G", "--groupbase", action="store_true",
-                          dest="groupbase", help="hide group source prefix " \
+                          dest="groupbase", help="hide group source prefix "
                           "(always \"@groupname\")")
         optgrp.add_option("-S", "--separator", action="store", dest="separator",
-                          default=' ', help="separator string to use when " \
-                          "expanding nodesets (default: ' ')")
+                          default=' ', help="separator string to use when "
+                                            "expanding nodesets (default: ' ')")
+        optgrp.add_option("-O", "--output-format", action="store",
+                          dest="output_format", metavar='FORMAT', default='%s',
+                          help="output format (default: '%s')")
         optgrp.add_option("-I", "--slice", action="store",
                           dest="slice_rangeset",
                           help="return sliced off result", type="string")
-        optgrp.add_option("--split", action="store", dest="maxsplit", 
+        optgrp.add_option("--split", action="store", dest="maxsplit",
                           help="split result into a number of subsets",
                           type="int")
         optgrp.add_option("--contiguous", action="store_true",
-                          dest="contiguous", help="split result into " \
-                          "contiguous subsets")
+                          dest="contiguous", help="split result into "
+                                                  "contiguous subsets")
+        optgrp.add_option("--axis", action="store", dest="axis",
+                          metavar="RANGESET", help="fold along these axis only "
+                                                   "(axis 1..n for nD nodeset)")
         self.add_option_group(optgrp)
-
-
