@@ -1,5 +1,5 @@
 #
-# Copyright CEA/DAM/DIF (2007, 2008, 2009, 2010, 2011)
+# Copyright CEA/DAM/DIF (2007-2015)
 #  Contributor: Stephane THIELL <stephane.thiell@cea.fr>
 #
 # This file is part of the ClusterShell library.
@@ -40,8 +40,16 @@ in response to incoming events (from workers, timers, etc.).
 import errno
 import heapq
 import logging
+import sys
 import time
+import traceback
 
+# Engine client fd I/O event interest bits
+E_READ = 0x1
+E_WRITE = 0x2
+
+# Define epsilon value for time float arithmetic operations
+EPSILON = 1.0e-3
 
 class EngineException(Exception):
     """
@@ -122,7 +130,7 @@ class EngineBaseTimer:
         Returns a boolean value that indicates whether an EngineTimer
         object is valid and able to fire.
         """
-        return self._engine != None
+        return self._engine is not None
 
     def set_nextfire(self, fire_delay, interval=-1):
         """
@@ -171,7 +179,7 @@ class EngineTimer(EngineBaseTimer):
     def __init__(self, fire_delay, interval, autoclose, handler):
         EngineBaseTimer.__init__(self, fire_delay, interval, autoclose)
         self.eh = handler
-        assert self.eh != None, "An event handler is needed for timer."
+        assert self.eh is not None, "An event handler is needed for timer."
 
     def _fire(self):
         self.eh.ev_timer(self)
@@ -187,29 +195,33 @@ class _EngineTimerQ:
             self.client = client
             self.client._timercase = self
             # arm timer (first time)
-            assert self.client.fire_delay > 0
+            assert self.client.fire_delay > -EPSILON
             self.fire_date = self.client.fire_delay + time.time()
 
         def __cmp__(self, other):
             return cmp(self.fire_date, other.fire_date)
 
         def arm(self, client):
-            assert client != None
+            assert client is not None
             self.client = client
             self.client._timercase = self
             # setup next firing date
             time_current = time.time()
-            if self.client.fire_delay > 0:
+            if self.client.fire_delay > -EPSILON:
                 self.fire_date = self.client.fire_delay + time_current
             else:
                 interval = float(self.client.interval)
                 assert interval > 0
+                # Keep it simple: increase fire_date by interval even if
+                # fire_date stays in the past, as in that case it's going to
+                # fire again at next runloop anyway.
                 self.fire_date += interval
-                # If the firing time is delayed so far that it passes one
-                # or more of the scheduled firing times, reschedule the
-                # timer for the next scheduled firing time in the future.
-                while self.fire_date < time_current:
-                    self.fire_date += interval
+                # Just print a debug message that could help detect issues
+                # coming from a long-running timer handler.
+                if self.fire_date < time_current:
+                    logging.getLogger(__name__).debug(
+                        "Warning: passed interval time for %r (long running "
+                        "event handler?)", self.client)
 
         def disarm(self):
             client = self.client
@@ -218,8 +230,8 @@ class _EngineTimerQ:
             return client
 
         def armed(self):
-            return self.client != None
-            
+            return self.client is not None
+
 
     def __init__(self, engine):
         """
@@ -240,7 +252,7 @@ class _EngineTimerQ:
         Insert and arm a client's timer.
         """
         # arm only if fire is set
-        if client.fire_delay > 0:
+        if client.fire_delay > -EPSILON:
             heapq.heappush(self.timers, _EngineTimerQ._EngineTimerCase(client))
             self.armed_count += 1
             if not client.autoclose:
@@ -262,8 +274,8 @@ class _EngineTimerQ:
         """
         if not client._timercase:
             # if timer is being fire, invalidate its values
-            client.fire_delay = 0
-            client.interval = 0
+            client.fire_delay = -1.0
+            client.interval = -1.0
             return
 
         if self.armed_count <= 0:
@@ -281,26 +293,41 @@ class _EngineTimerQ:
         while len(self.timers) > 0 and not self.timers[0].armed():
             heapq.heappop(self.timers)
 
-    def fire(self):
+    def fire_expired(self):
         """
-        Remove the smallest timer from the queue and fire its associated client.
-        Raise IndexError if the queue is empty.
+        Remove expired timers from the queue and fire associated clients.
         """
         self._dequeue_disarmed()
 
-        timercase = heapq.heappop(self.timers)
-        client = timercase.disarm()
-        
-        client.fire_delay = 0
-        client._fire()
+        # Build a queue of expired timercases. Any expired (and still armed)
+        # timer is fired, but only once per call.
+        expired_timercases = []
+        now = time.time()
+        while self.timers and (self.timers[0].fire_date - now) <= EPSILON:
+            expired_timercases.append(heapq.heappop(self.timers))
+            self._dequeue_disarmed()
 
-        if client.fire_delay > 0 or client.interval > 0:
-            timercase.arm(client)
-            heapq.heappush(self.timers, timercase)
-        else:
-            self.armed_count -= 1
-            if not client.autoclose:
-                self._engine.evlooprefcnt -= 1
+        for timercase in expired_timercases:
+            # Be careful to recheck and skip any disarmed timers (eg. timer
+            # could be invalidated from another timer's event handler)
+            if not timercase.armed():
+                continue
+
+            # Disarm timer
+            client = timercase.disarm()
+
+            # Fire timer
+            client.fire_delay = -1.0
+            client._fire()
+
+            # Rearm it if needed - Note: fire=0 is valid, interval=0 is not
+            if client.fire_delay >= -EPSILON or client.interval > EPSILON:
+                timercase.arm(client)
+                heapq.heappush(self.timers, timercase)
+            else:
+                self.armed_count -= 1
+                if not client.autoclose:
+                    self._engine.evlooprefcnt -= 1
 
     def nextfire_delay(self):
         """
@@ -312,14 +339,6 @@ class _EngineTimerQ:
 
         return -1
 
-    def expired(self):
-        """
-        Has a timer expired?
-        """
-        self._dequeue_disarmed()
-        return len(self.timers) > 0 and \
-            (self.timers[0].fire_date - time.time()) <= 1e-2
-
     def clear(self):
         """
         Stop and clear all timers.
@@ -334,22 +353,17 @@ class _EngineTimerQ:
 
 class Engine:
     """
-    Interface for ClusterShell engine. Subclasses have to implement a runloop
-    listening for client events.
-    """
+    Base class for ClusterShell Engines.
 
-    # Engine client I/O event interest bits
-    E_READ = 0x1
-    E_ERROR = 0x2
-    E_WRITE = 0x4
-    E_ANY = E_READ | E_ERROR | E_WRITE
+    Subclasses have to implement a runloop listening for client events.
+    Subclasses that override other than "pure virtual methods" should call
+    corresponding base class methods.
+    """
 
     identifier = "(none)"
 
     def __init__(self, info):
-        """
-        Initialize base class.
-        """
+        """Initialize base class."""
         # take a reference on info dict
         self.info = info
 
@@ -364,7 +378,7 @@ class Engine:
         self.reg_clients = 0
 
         # keep track of registered file descriptors in a dict where keys
-        # are fileno and values are clients
+        # are fileno and values are (EngineClient, EngineClientStream) tuples
         self.reg_clifds = {}
 
         # Current loop iteration counter. It is the number of performed engine
@@ -372,8 +386,8 @@ class Engine:
         # safely process FDs by chunk and re-use FDs (see Engine._fd2client).
         self._current_loopcnt = 0
 
-        # Current client being processed
-        self._current_client = None
+        # Current stream being processed
+        self._current_stream = None
 
         # timer queue to handle both timers and clients timeout
         self.timerq = _EngineTimerQ(self)
@@ -387,10 +401,12 @@ class Engine:
         # runloop-has-exited flag
         self._exited = False
 
+    def release(self):
+        """Release engine-specific resources."""
+        pass
+
     def clients(self):
-        """
-        Get a copy of clients set.
-        """
+        """Get a copy of clients set."""
         return self._clients.copy()
 
     def ports(self):
@@ -400,20 +416,17 @@ class Engine:
         return self._ports.copy()
 
     def _fd2client(self, fd):
-        client, fdev = self.reg_clifds.get(fd, (None, None))
+        client, stream = self.reg_clifds.get(fd, (None, None))
         if client:
             if client._reg_epoch < self._current_loopcnt:
-                return client, fdev
+                return client, stream
             else:
                 self._debug("ENGINE _fd2client: ignoring just re-used FD %d" \
-                            % fd)
+                            % stream.fd)
         return (None, None)
 
     def add(self, client):
-        """
-        Add a client to engine. Subclasses that override this method
-        should call base class method.
-        """
+        """Add a client to engine."""
         # bind to engine
         client._set_engine(self)
 
@@ -431,21 +444,19 @@ class Engine:
             elif self.info["fanout"] > self.reg_clients:
                 self.register(client._start())
 
-    def _remove(self, client, abort, did_timeout=False, force=False):
-        """
-        Remove a client from engine (subroutine).
-        """
+    def _remove(self, client, abort, did_timeout=False):
+        """Remove a client from engine (subroutine)."""
         # be careful to also remove ports when engine has not started yet
         if client.registered or not client.delayable:
             if client.registered:
                 self.unregister(client)
             # care should be taken to ensure correct closing flags
-            client._close(abort=abort, flush=not force, timeout=did_timeout)
+            client._close(abort=abort, timeout=did_timeout)
 
     def remove(self, client, abort=False, did_timeout=False):
         """
-        Remove a client from engine. Subclasses that override this
-        method should call base class method.
+        Remove a client from engine. Does NOT aim to flush individual stream
+        read buffers.
         """
         self._debug("REMOVE %s" % client)
         if client.delayable:
@@ -454,11 +465,26 @@ class Engine:
             self._ports.remove(client)
         self._remove(client, abort, did_timeout)
         self.start_all()
-    
+
+    def remove_stream(self, client, stream):
+        """
+        Regular way to remove a client stream from engine, performing
+        needed read flush as needed. If no more retainable stream
+        remains for this client, this method automatically removes the
+        entire client from engine.
+        """
+        self.unregister_stream(client, stream)
+        # _close_stream() will flush pending read buffers so may generate events
+        client._close_stream(stream.name)
+        # client may have been removed by previous events, if not check whether
+        # some retained streams still remain
+        if client in self._clients and not client.streams.retained():
+            self.remove(client)
+
     def clear(self, did_timeout=False, clear_ports=False):
         """
-        Remove all clients. Subclasses that override this method should
-        call base class method.
+        Remove all clients. Does not flush read buffers.
+        Subclasses that override this method should call base class method.
         """
         all_clients = [self._clients]
         if clear_ports:
@@ -467,7 +493,7 @@ class Engine:
         for clients in all_clients:
             while len(clients) > 0:
                 client = clients.pop()
-                self._remove(client, True, did_timeout, force=True)
+                self._remove(client, True, did_timeout)
 
     def register(self, client):
         """
@@ -477,121 +503,80 @@ class Engine:
         assert client in self._clients or client in self._ports
         assert not client.registered
 
-        efd = client.fd_error
-        rfd = client.fd_reader
-        wfd = client.fd_writer
-        assert rfd is not None or wfd is not None
+        self._debug("REG %s (%s)(autoclose=%s)" % \
+                (client.__class__.__name__, client.streams,
+                 client.autoclose))
 
-        self._debug("REG %s(e%s,r%s,w%s)(autoclose=%s)" % \
-                (client.__class__.__name__, efd, rfd, wfd,
-                    client.autoclose))
-
-        client._events = 0
         client.registered = True
         client._reg_epoch = self._current_loopcnt
 
         if client.delayable:
             self.reg_clients += 1
 
-        if client.autoclose:
-            refcnt_inc = 0
-        else:
-            refcnt_inc = 1
-
-        if efd != None:
-            self.reg_clifds[efd] = client, Engine.E_ERROR
-            client._events |= Engine.E_ERROR
-            self.evlooprefcnt += refcnt_inc
-            self._register_specific(efd, Engine.E_ERROR)
-        if rfd != None:
-            self.reg_clifds[rfd] = client, Engine.E_READ
-            client._events |= Engine.E_READ
-            self.evlooprefcnt += refcnt_inc
-            self._register_specific(rfd, Engine.E_READ)
-        if wfd != None:
-            self.reg_clifds[wfd] = client, Engine.E_WRITE
-            client._events |= Engine.E_WRITE
-            self.evlooprefcnt += refcnt_inc
-            self._register_specific(wfd, Engine.E_WRITE)
-
-        client._new_events = client._events
+        # set interest event bits...
+        for streams, ievent in ((client.streams.active_readers, E_READ),
+                                (client.streams.active_writers, E_WRITE)):
+            for stream in streams():
+                self.reg_clifds[stream.fd] = client, stream
+                stream.events |= ievent
+                if not client.autoclose:
+                    self.evlooprefcnt += 1
+                self._register_specific(stream.fd, ievent)
 
         # start timeout timer
         self.timerq.schedule(client)
 
-    def unregister_writer(self, client):
-        self._debug("UNREG WRITER r%s,w%s" % (client.reader_fileno(), \
-            client.writer_fileno()))
-        if client.autoclose:
-            refcnt_inc = 0
-        else:
-            refcnt_inc = 1
-
-        wfd = client.fd_writer
-        if wfd != None:
-            self._unregister_specific(wfd, client._events & Engine.E_WRITE)
-            client._events &= ~Engine.E_WRITE
-            del self.reg_clifds[wfd]
-            self.evlooprefcnt -= refcnt_inc
+    def unregister_stream(self, client, stream):
+        """Unregister a stream from a client."""
+        self._debug("UNREG_STREAM stream=%s" % stream)
+        assert stream is not None and stream.fd is not None
+        assert stream.fd in self.reg_clifds, \
+            "stream fd %d not registered" % stream.fd
+        assert client.registered
+        self._unregister_specific(stream.fd, stream.events & stream.evmask)
+        self._debug("UNREG_STREAM unregistering stream fd %d (%d)" % \
+                    (stream.fd, len(client.streams)))
+        stream.events &= ~stream.evmask
+        del self.reg_clifds[stream.fd]
+        if not client.autoclose:
+            self.evlooprefcnt -= 1
 
     def unregister(self, client):
-        """
-        Unregister a client. Subclasses that override this method should
-        call base class method.
-        """
+        """Unregister a client"""
         # sanity check
         assert client.registered
-        self._debug("UNREG %s (r%s,e%s,w%s)" % (client.__class__.__name__,
-            client.reader_fileno(), client.error_fileno(),
-            client.writer_fileno()))
-        
+        self._debug("UNREG %s (%s)" % (client.__class__.__name__, \
+                client.streams))
+
         # remove timeout timer
         self.timerq.invalidate(client)
-        
-        if client.autoclose:
-            refcnt_inc = 0
-        else:
-            refcnt_inc = 1
-            
-        # clear interest events
-        efd = client.fd_error
-        if efd is not None:
-            self._unregister_specific(efd, client._events & Engine.E_ERROR)
-            client._events &= ~Engine.E_ERROR
-            del self.reg_clifds[efd]
-            self.evlooprefcnt -= refcnt_inc
 
-        rfd = client.fd_reader
-        if rfd is not None:
-            self._unregister_specific(rfd, client._events & Engine.E_READ)
-            client._events &= ~Engine.E_READ
-            del self.reg_clifds[rfd]
-            self.evlooprefcnt -= refcnt_inc
+        # clear interest events...
+        for streams, ievent in ((client.streams.active_readers, E_READ),
+                                (client.streams.active_writers, E_WRITE)):
+            for stream in streams():
+                if stream.fd in self.reg_clifds:
+                    self._unregister_specific(stream.fd, stream.events & ievent)
+                    stream.events &= ~ievent
+                    del self.reg_clifds[stream.fd]
+                    if not client.autoclose:
+                        self.evlooprefcnt -= 1
 
-        wfd = client.fd_writer
-        if wfd is not None:
-            self._unregister_specific(wfd, client._events & Engine.E_WRITE)
-            client._events &= ~Engine.E_WRITE
-            del self.reg_clifds[wfd]
-            self.evlooprefcnt -= refcnt_inc
-
-        client._new_events = 0
         client.registered = False
         if client.delayable:
             self.reg_clients -= 1
 
-    def modify(self, client, setmask, clearmask):
-        """
-        Modify the next loop interest events bitset for a client.
-        """
-        self._debug("MODEV set:0x%x clear:0x%x %s" % (setmask, clearmask,
-                                                      client))
-        client._new_events &= ~clearmask
-        client._new_events |= setmask
+    def modify(self, client, sname, setmask, clearmask):
+        """Modify the next loop interest events bitset for a client stream."""
+        self._debug("MODEV set:0x%x clear:0x%x %s (%s)" % (setmask, clearmask,
+                                                           client, sname))
+        stream = client.streams[sname]
+        stream.new_events &= ~clearmask
+        stream.new_events |= setmask
 
-        if self._current_client is not client:
-            # modifying a non processing client, apply new_events now
-            self.set_events(client, client._new_events)
+        if self._current_stream is not stream:
+            # modifying a non processing stream, apply new_events now
+            self.set_events(client, stream)
 
     def _register_specific(self, fd, event):
         """Engine-specific register fd for event method."""
@@ -604,101 +589,62 @@ class Engine:
     def _modify_specific(self, fd, event, setvalue):
         """Engine-specific modify fd for event method."""
         raise NotImplementedError("Derived classes must implement.")
-        
-    def set_events(self, client, new_events):
-        """
-        Set the active interest events bitset for a client.
-        """
-        self._debug("SETEV new_events:0x%x events:0x%x %s" % (new_events,
-            client._events, client))
+
+    def set_events(self, client, stream):
+        """Set the active interest events bitset for a client stream."""
+        self._debug("SETEV new_events:0x%x events:0x%x for %s[%s]" % \
+            (stream.new_events, stream.events, client, stream.name))
 
         if not client.registered:
             logging.getLogger(__name__).debug( \
                 "set_events: client %s not registered" % self)
             return
 
-        chgbits = new_events ^ client._events
+        chgbits = stream.new_events ^ stream.events
         if chgbits == 0:
             return
 
         # configure interest events as appropriate
-        efd = client.fd_error
-        if efd is not None:
-            if chgbits & Engine.E_ERROR:
-                status = new_events & Engine.E_ERROR
-                self._modify_specific(efd, Engine.E_ERROR, status)
+        for interest in (E_READ, E_WRITE):
+            if chgbits & interest:
+                assert stream.evmask & interest
+                status = stream.new_events & interest
+                self._modify_specific(stream.fd, interest, status)
                 if status:
-                    client._events |= Engine.E_ERROR
+                    stream.events |= interest
                 else:
-                    client._events &= ~Engine.E_ERROR
+                    stream.events &= ~interest
 
-        rfd = client.fd_reader
-        if rfd is not None:
-            if chgbits & Engine.E_READ:
-                status = new_events & Engine.E_READ
-                self._modify_specific(rfd, Engine.E_READ, status)
-                if status:
-                    client._events |= Engine.E_READ
-                else:
-                    client._events &= ~Engine.E_READ
+        stream.new_events = stream.events
 
-        wfd = client.fd_writer
-        if wfd is not None:
-            if chgbits & Engine.E_WRITE:
-                status = new_events & Engine.E_WRITE
-                self._modify_specific(wfd, Engine.E_WRITE, status)
-                if status:
-                    client._events |= Engine.E_WRITE
-                else:
-                    client._events &= ~Engine.E_WRITE
-
-        client._new_events = client._events
-
-    def set_reading(self, client):
-        """
-        Set client reading state.
-        """
+    def set_reading(self, client, sname):
+        """Set client reading state."""
         # listen for readable events
-        self.modify(client, Engine.E_READ, 0)
+        self.modify(client, sname, E_READ, 0)
 
-    def set_reading_error(self, client):
-        """
-        Set client reading error state.
-        """
-        # listen for readable events
-        self.modify(client, Engine.E_ERROR, 0)
-
-    def set_writing(self, client):
-        """
-        Set client writing state.
-        """
+    def set_writing(self, client, sname):
+        """Set client writing state."""
         # listen for writable events
-        self.modify(client, Engine.E_WRITE, 0)
+        self.modify(client, sname, E_WRITE, 0)
 
     def add_timer(self, timer):
-        """
-        Add engine timer.
-        """
+        """Add a timer instance to engine."""
         timer._set_engine(self)
         self.timerq.schedule(timer)
 
     def remove_timer(self, timer):
-        """
-        Remove engine timer.
-        """
+        """Remove engine timer from engine."""
         self.timerq.invalidate(timer)
 
     def fire_timers(self):
-        """
-        Fire expired timers for processing.
-        """
-        while self.timerq.expired():
-            self.timerq.fire()
+        """Fire expired timers for processing."""
+        # Only fire timers if runloop is still retained
+        if self.evlooprefcnt > 0:
+            # Fire once any expired timers
+            self.timerq.fire_expired()
 
     def start_ports(self):
-        """
-        Start and register all port clients.
-        """
+        """Start and register all port clients."""
         # Ports are special, non-delayable engine clients
         for port in self._ports:
             if not port.registered:
@@ -707,7 +653,8 @@ class Engine:
 
     def start_all(self):
         """
-        Start and register all other possible clients, in respect of task fanout.
+        Start and register all other possible clients, in respect of task
+        fanout.
         """
         # Get current fanout value
         fanout = self.info["fanout"]
@@ -722,32 +669,44 @@ class Engine:
                 self.register(client._start())
                 if fanout <= self.reg_clients:
                     break
-    
+
     def run(self, timeout):
-        """
-        Run engine in calling thread.
-        """
+        """Run engine in calling thread."""
         # change to running state
         if self.running:
             raise EngineAlreadyRunningError()
-        self.running = True
-
-        # start port clients
-        self.start_ports()
-
-        # peek in ports for early pending messages
-        self.snoop_ports()
-
-        # start all other clients
-        self.start_all()
 
         # note: try-except-finally not supported before python 2.5
         try:
+            self.running = True
             try:
+                # start port clients
+                self.start_ports()
+                # peek in ports for early pending messages
+                self.snoop_ports()
+                # start all other clients
+                self.start_all()
+                # run loop until all clients and timers are removed
                 self.runloop(timeout)
-            except Exception, e:
-                # any exceptions invalidate clients
-                self.clear(isinstance(e, EngineTimeoutException))
+            except EngineTimeoutException:
+                self.clear(did_timeout=True)
+                raise
+            except: # MUST use BaseException as soon as possible (py2.5+)
+                # The game is over.
+                exc_t, exc_val, exc_tb = sys.exc_info()
+                try:
+                    # Close Engine clients
+                    self.clear()
+                except:
+                    # self.clear() may still generate termination events that
+                    # may raises exceptions, overriding the other one above.
+                    # In the future, we should block new user events to avoid
+                    # that. Also, such cases could be better handled with
+                    # BaseException. For now, print a backtrace in debug to
+                    # help detect the problem.
+                    tbexc = traceback.format_exception(exc_t, exc_val, exc_tb)
+                    logging.getLogger(__name__).debug(''.join(tbexc))
+                    raise
                 raise
         finally:
             # cleanup
@@ -757,46 +716,37 @@ class Engine:
     def snoop_ports(self):
         """
         Peek in ports for possible early pending messages.
-        This method simply tries to read port pipes in non-
-        blocking mode.
+        This method simply tries to read port pipes in non-blocking mode.
         """
         # make a copy so that early messages on installed ports may
         # lead to new ports
         ports = self._ports.copy()
         for port in ports:
             try:
-                port._handle_read()
-            except (IOError, OSError), (err, strerr):
-                if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
+                port._handle_read('in')
+            except (IOError, OSError), ex:
+                if ex.errno in (errno.EAGAIN, errno.EWOULDBLOCK):
                     # no pending message
                     return
                 # raise any other error
                 raise
 
     def runloop(self, timeout):
-        """
-        Engine specific run loop. Derived classes must implement.
-        """
+        """Engine specific run loop. Derived classes must implement."""
         raise NotImplementedError("Derived classes must implement.")
 
     def abort(self, kill):
-        """
-        Abort runloop.
-        """
+        """Abort runloop."""
         if self.running:
             raise EngineAbortException(kill)
 
         self.clear(clear_ports=kill)
 
     def exited(self):
-        """
-        Returns True if the engine has exited the runloop once.
-        """
+        """Returns True if the engine has exited the runloop once."""
         return not self.running and self._exited
 
     def _debug(self, s):
-        # library engine debugging hook
-        #import sys
-        #print >>sys.stderr, s
+        """library engine debugging hook"""
+        #logging.getLogger(__name__).debug(s)
         pass
-
