@@ -1,6 +1,6 @@
 #
-# Copyright CEA/DAM/DIF (2007, 2008, 2009, 2010, 2011, 2012)
-#  Contributor: Stephane THIELL <stephane.thiell@cea.fr>
+# Copyright CEA/DAM/DIF (2007-2015)
+#  Contributor: Stephane THIELL <sthiell@stanford.edu>
 #
 # This file is part of the ClusterShell library.
 #
@@ -48,9 +48,8 @@ Please see first:
   - ClusterShell.Task
 """
 
-__version__ = '1.6'
+__version__ = '1.7'
 __version_info__ = tuple([ int(_n) for _n in __version__.split('.')])
-__date__    = '2012/04/08'
-__author__  = 'Stephane Thiell <stephane.thiell@cea.fr>'
-__url__     = 'http://cea-hpc.github.com/clustershell/'
-
+__date__    = '2015/11/10'
+__author__  = 'Stephane Thiell <sthiell@stanford.edu>'
+__url__     = 'http://clustershell.readthedocs.org/'
